//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Files/DatafilesSet.cpp
//! @brief     Implements class DatafilesSet.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Files/DatafilesSet.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {
namespace Tag {

const QString RealData("RealData");
const QString CurrentIndex("CurrentIndex");
const QString SelectedRank("SelectedRank");

} // namespace Tag
} // namespace

DatafilesSet::DatafilesSet()
{
    setObjectName("DatafilesSet");
}

DatafilesSet::~DatafilesSet() = default;

QVector<DataItem*> DatafilesSet::dataItems() const
{
    QVector<DataItem*> result;
    for (auto* dfile_item : dfileItems())
        if (auto* data_item = dfile_item->dataItem())
            result.push_back(data_item);

    return result;
}

void DatafilesSet::clear()
{
    for (auto* dfile_item : dfileItems())
        removeDatafileItem(dfile_item);
}

void DatafilesSet::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // real items
    for (auto* dfile_item : dfileItems()) {
        w->writeStartElement(Tag::RealData);
        XML::writeAttribute(w, XML::Attrib::name, dfile_item->realItemName());
        dfile_item->writeTo(w);
        w->writeEndElement();
    }

    // selected index
    w->writeStartElement(Tag::CurrentIndex);
    XML::writeAttribute(w, XML::Attrib::value, m_current_index);
    w->writeEndElement();

    // selected rank
    w->writeStartElement(Tag::SelectedRank);
    XML::writeAttribute(w, XML::Attrib::value, m_selected_rank);
    w->writeEndElement();
}

void DatafilesSet::readFrom(QXmlStreamReader* r)
{
    clear();

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // real item
        if (tag == Tag::RealData) {
            auto* item = new DatafileItem;
            item->readFrom(r);
            m_dfile_items.emplace_back(item);
            XML::gotoEndElementOfTag(r, tag);

            // selected index
        } else if (tag == Tag::CurrentIndex) {
            XML::readAttribute(r, XML::Attrib::value, &m_current_index);
            XML::gotoEndElementOfTag(r, tag);

            // selected rank
        } else if (tag == Tag::SelectedRank) {
            XML::readAttribute(r, XML::Attrib::value, &m_selected_rank);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }

    if (r->hasError())
        throw std::runtime_error(r->errorString().toLatin1());
}

void DatafilesSet::writeDatafiles(const QString& projectDir)
{
    for (const auto* dfile_item : dfileItems())
        dfile_item->writeDatafiles(projectDir);

    dataFilesCleaner.cleanOldFiles(projectDir, dataItems());
}

void DatafilesSet::readDatafiles(const QString& projectDir, MessageService* messageService)
{
    for (auto* dfile_item : dfileItems())
        dfile_item->readDatafiles(projectDir, messageService);

    dataFilesCleaner.recollectDataNames(dataItems());
}

void DatafilesSet::insertDatafileItem(DatafileItem* dfi)
{
    m_dfile_items.emplace_back(dfi);
}

QVector<DatafileItem*> DatafilesSet::dfileItems() const
{
    QVector<DatafileItem*> items(m_dfile_items.size());
    for (int i = 0; i < items.size(); i++)
        items[i] = m_dfile_items[i].get();
    return items;
}


QVector<DatafileItem*> DatafilesSet::dfileItems(int rank) const
{
    return rank == 1 ? realItems1D() : realItems2D();
}

QVector<DatafileItem*> DatafilesSet::realItems1D() const
{
    QVector<DatafileItem*> vec_1d;

    for (DatafileItem* dfile_item : dfileItems())
        if (dfile_item->rank() == 1)
            vec_1d.push_back(dfile_item);

    return vec_1d;
}

QVector<DatafileItem*> DatafilesSet::realItems2D() const
{
    QVector<DatafileItem*> vec_2d;

    for (DatafileItem* dfile_item : dfileItems())
        if (dfile_item->rank() == 2)
            vec_2d.push_back(dfile_item);

    return vec_2d;
}

void DatafilesSet::removeDatafileItem(DatafileItem* dfile_item)
{
    ASSERT(dfile_item);
    int index = dfileItems().indexOf(dfile_item);
    m_dfile_items.erase(m_dfile_items.begin() + index);
}

QStringList DatafilesSet::realItemNames() const
{
    QStringList result;
    for (DatafileItem* dfile_item : dfileItems())
        result.append(dfile_item->realItemName());
    return result;
}
