//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/JobItem.cpp
//! @brief     Implements class JobItem.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/JobItem.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/Scale.h"
#include "Device/Data/Datafield.h"
#include "Device/Detector/IDetector.h"
#include "GUI/Model/Axis/AmplitudeAxisItem.h"
#include "GUI/Model/Data/Data1DItem.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Detector/DetectorItem.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/BatchInfo.h"
#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Job/ParameterTreeBuilder.h"
#include "GUI/Model/Job/ParameterTreeItems.h"
#include "GUI/Model/Mask/MaskUnitsConverter.h"
#include "GUI/Model/Mask/MasksSet.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Model/ToCore/SimulationToCore.h"
#include "GUI/Support/Data/JobStatus.h"
#include "GUI/Support/Data/JobWorker.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Support/Util/Path.h"
#include "GUI/Support/XML/Backup.h"
#include "Sim/Simulation/ISimulation.h"

namespace {
namespace Tag {

const QString Activity("Activity");
const QString BatchInfo("BatchInfo");
const QString DatafileItem("DatafileItem");
const QString FitSuite("FitSuite");
const QString Instrument("Instrument");
const QString ParameterContainer("ParameterContainer");
const QString Sample("Sample");
const QString SimulatedData("SimulatedData");
const QString SimulationOptions("SimulationOptions");

} // namespace Tag

DataItem* newDataItem(size_t rank)
{
    if (rank == 1)
        return new Data1DItem();
    else if (rank == 2)
        return new Data2DItem();
    else
        ASSERT_NEVER;
}

} // namespace

JobItem::JobItem()
    : m_simulation_options_item(std::make_unique<SimulationOptionsItem>())
    , m_parameter_container(std::make_unique<ParameterContainerItem>())
    , m_sample_item(std::make_unique<SampleItem>())
    , m_batch_info(std::make_unique<BatchInfo>())
    , m_fit_suite_item(std::make_unique<FitSuiteItem>())
{
}

JobItem::JobItem(const SampleItem* sampleItem, const InstrumentItem* instrumentItem,
                 const DatafileItem* dfile_item, const SimulationOptionsItem* optionItem)
    : m_simulation_options_item(std::make_unique<SimulationOptionsItem>(*optionItem))
    , m_parameter_container(std::make_unique<ParameterContainerItem>())
    , m_sample_item(sampleItem->clone())
    , m_batch_info(std::make_unique<BatchInfo>())
    , m_fit_suite_item(std::make_unique<FitSuiteItem>())
{
    ASSERT(sampleItem);
    ASSERT(instrumentItem);

    batchInfo()->setIdentifier(QUuid::createUuid().toString());

    copyInstrumentIntoJob(instrumentItem);

    createParameterTree();
    parameterContainerItem()->addBackupValues("Initial state");

    createSimulatedDataItem();

    if (dfile_item) {
        copyDatafileItemIntoJob(dfile_item);
        adjustReaDataToJobInstrument();

        createDiffDataItem();
    }
}

JobItem::~JobItem() = default;

void JobItem::setFailed()
{
    if (DataItem* dataItem = simulatedDataItem()) {
        if (Datafield* df = dataItem->p_field())
            df->setAllTo(0.0);
        emit dataItem->datafieldChanged();
    }
    batchInfo()->setStatus(JobStatus::Failed);
}

size_t JobItem::rank() const
{
    if (instrumentItem()->is<SpecularInstrumentItem>())
        return 1;
    else if (instrumentItem()->is<GISASInstrumentItem>()
             || instrumentItem()->is<OffspecInstrumentItem>()
             || instrumentItem()->is<DepthprobeInstrumentItem>())
        return 2;
    else
        ASSERT_NEVER;
}

void JobItem::createParameterTree()
{
    ParameterTreeBuilder(this).build();
}

void JobItem::copyInstrumentIntoJob(const InstrumentItem* instrument)
{
    m_instrument.setCurrentItem(instrument->createItemCopy());
    m_instrument.currentItem()->setId(QUuid::createUuid().toString());
}

void JobItem::createSimulatedDataItem()
{
    ASSERT(!simulatedDataItem());
    m_simulated_data_item.reset(::newDataItem(rank()));
}

Data2DItem* JobItem::data2DItem()
{
    return dynamic_cast<Data2DItem*>(m_simulated_data_item.get());
}

const DataItem* JobItem::createDiffDataItem()
{
    ASSERT(!diffDataItem());
    m_diff_data_item.reset(::newDataItem(rank()));

    ASSERT(m_dfile_item);
    // use the same axes as simulated item (same as external data item if loading pre-22.0 project)
    const DataItem* source = m_simulated_data_item ? simulatedDataItem() : m_dfile_item->dataItem();
    GUI::Util::copyContents(source->axItemX(), m_diff_data_item->axItemX());
    if (rank() == 2)
        GUI::Util::copyContents(source->axItemY(), m_diff_data_item->axItemY());

    if (auto* spec_diff = dynamic_cast<Data1DItem*>(diffDataItem()))
        spec_diff->setDiffPlotStyle();

    if (auto* intensity_diff = dynamic_cast<Data2DItem*>(diffDataItem()))
        intensity_diff->setCurrentGradient(m_dfile_item->data2DItem()->currentGradient());

    return m_diff_data_item.get();
}

void JobItem::copyDatafileItemIntoJob(const DatafileItem* source)
{
    ASSERT(!dfileItem());
    ASSERT(source->rank() == rank());
    m_dfile_item.reset(source->clone());

    if (rank() == 1) {
        m_dfile_item->data1DItem()->setRealPlotStyle();
    } else if (rank() == 2) {
        // use color scheme from loaded data
        ASSERT(data2DItem());
        data2DItem()->setCurrentGradient(m_dfile_item->data2DItem()->currentGradient());
    }
}

void JobItem::adjustReaDataToJobInstrument()
{
    // update stored instrument id without updating rest of real data
    m_dfile_item->setInstrumentId(instrumentItem()->id());

    if (instrumentItem()->is<SpecularInstrumentItem>()) {

        // coords of loaded data --> instrumental --> representation coords
        const Datafield* old = m_dfile_item->data1DItem()->c_field();
        Frame frame = instrumentItem()->makeFrame();
        m_dfile_item->data1DItem()->setDatafield(
            {frame.clone(), old->flatVector(), old->errorSigmas()});

    } else if (instrumentItem()->is<GISASInstrumentItem>()) {

        // Before copying masks on board of detector we need to ensure that
        // mask items are represented in proper coordinates.
        // Usual unit/coord conversions touch only Frame of Datafield, but not masks.
        convertMaskCoords();

        importMasksFromDatafileItem(); // Copy masks & ROI from DatafileItem on board of instrument.
        cropRealData();                // Crop DatafileItem to the region of interest.
    }
}

void JobItem::importMasksFromDatafileItem()
{
    if (auto* iiI = dynamic_cast<GISASInstrumentItem*>(instrumentItem()))
        if (const auto* d2i = dfileItem()->data2DItem())
            if (const auto* container = d2i->masksSet())
                iiI->detectorItem()->importMasks(container);
}

void JobItem::convertMaskCoords()
{
    const Datafield* dfield = m_dfile_item->data2DItem()->c_field();
    ASSERT(dfield);
    MaskUnitsConverter mask_converter(dfield->frame());

    // Masks conversion can be done in two steps: first from original coords to bins,
    // then from bins to target coords.
    // If 'IDetector::addBinMask' is used by detector, only "orig coords --> bins" is needed.
    if (const auto* container = m_dfile_item->data2DItem()->masksSet())
        for (auto* maskItem : container->modifiableMaskItems())
            mask_converter.convertToNbins(maskItem);
}

//! Crops DatafileItem to the region of interest.
void JobItem::cropRealData()
{
    auto* iiI = dynamic_cast<GISASInstrumentItem*>(instrumentItem());
    ASSERT(iiI);

    // Adjust real data to the size of region of interest
    Data2DItem* data_item = dfileItem()->data2DItem();

    std::unique_ptr<Datafield> origData(data_item->c_field()->clone());

    const Frame frame = iiI->makeFrame();

    // (re)create zero-valued Datafield.
    data_item->setDatafield({frame.clonedAxes()});

    std::unique_ptr<const IDetector> det = iiI->normalDetector();
    ASSERT(det->frame().rank() == 2);
    std::vector<size_t> ai = det->activeIndices();
    for (unsigned long i : ai)
        (*data_item->p_field())[i] = (*origData)[det->roiToFullIndex(i)];

    data_item->updateDataRange();
}

void JobItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(2));

    // simulation options
    w->writeStartElement(Tag::SimulationOptions);
    m_simulation_options_item->writeTo(w);
    w->writeEndElement();

    // instrument
    w->writeStartElement(Tag::Instrument);
    m_instrument.writeTo(w);
    w->writeEndElement();

    // sample
    w->writeStartElement(Tag::Sample);
    m_sample_item->writeTo(w);
    w->writeEndElement();

    // parameters
    w->writeStartElement(Tag::ParameterContainer);
    m_parameter_container->writeTo(w);
    w->writeEndElement();

    // batch info
    w->writeStartElement(Tag::BatchInfo);
    m_batch_info->writeTo(w);
    w->writeEndElement();

    // activity
    w->writeStartElement(Tag::Activity);
    XML::writeAttribute(w, XML::Attrib::value, m_activity);
    w->writeEndElement();

    // simulated data
    if (m_simulated_data_item) {
        w->writeStartElement(Tag::SimulatedData);
        m_simulated_data_item->writeTo(w);
        w->writeEndElement();
    }

    // real item
    if (m_dfile_item) {
        w->writeStartElement(Tag::DatafileItem);
        m_dfile_item->writeTo(w);
        w->writeEndElement();
    }

    // fit suite
    if (m_fit_suite_item) {
        w->writeStartElement(Tag::FitSuite);
        m_fit_suite_item->writeTo(w);
        w->writeEndElement();
    }
}

void JobItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // simulation options
        if (tag == Tag::SimulationOptions) {
            m_simulation_options_item->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // instrument
        } else if (tag == Tag::Instrument) {
            m_instrument.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // parameters
        } else if (tag == Tag::ParameterContainer) {
            createParameterTree();
            m_parameter_container->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // sample
        } else if (tag == Tag::Sample) {
            m_sample_item->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // activity
        } else if (tag == Tag::Activity) {
            XML::readAttribute(r, XML::Attrib::value, &m_activity);
            XML::gotoEndElementOfTag(r, tag);

            // comments
        } else if (tag == Tag::BatchInfo) {
            m_batch_info->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // simulated data
        } else if (tag == Tag::SimulatedData) {
            createSimulatedDataItem();
            m_simulated_data_item->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // real item
        } else if (tag == Tag::DatafileItem) {
            m_dfile_item = std::make_unique<DatafileItem>();
            m_dfile_item->readFrom(r);
            createDiffDataItem()->checkXYranges(m_dfile_item->dataItem());
            importMasksFromDatafileItem();
            XML::gotoEndElementOfTag(r, tag);

            // fit suite
        } else if (tag == Tag::FitSuite) {
            m_fit_suite_item->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

void JobItem::writeDatafiles(const QString& projectDir) const
{
    if (m_dfile_item)
        m_dfile_item->writeDatafiles(projectDir);

    if (m_simulated_data_item)
        m_simulated_data_item->saveDatafield(projectDir);
}

void JobItem::readDatafiles(const QString& projectDir, MessageService* messageService)
{
    QString realError, simError, errorMessage;

    if (m_dfile_item)
        realError = m_dfile_item->readDatafiles(projectDir, messageService);

    if (m_simulated_data_item)
        simError = m_simulated_data_item->loadDatafield(messageService, projectDir, rank());

    // Handling corrupted file on disk
    if (!realError.isEmpty() || !simError.isEmpty()) {
        errorMessage = QString("Load of the data from disk failed with:\n");

        if (!realError.isEmpty() && simError.isEmpty())
            errorMessage += QString("'%1'").arg(realError);
        else if (realError.isEmpty() && !simError.isEmpty())
            errorMessage += QString("'%1'").arg(simError);
        else
            errorMessage += QString("'%1',\n'%2'").arg(realError, simError);
    }

    // Handling previous crash of GUI during job run
    if (isRunning(batchInfo()->status())) {
        if (errorMessage.isEmpty())
            errorMessage = "Possible GUI crash while job was running";
        else
            errorMessage += "\n. Also possibly GUI crashed while job was running";
    }

    if (!errorMessage.isEmpty()) {
        batchInfo()->setComments(errorMessage);
        setFailed();
    }
}

//! Updates the name of file to store intensity data.
void JobItem::updateFileName()
{
    if (DataItem* item = simulatedDataItem())
        item->setFileName(GUI::Path::intensityDataFileName(batchInfo()->jobName(), "jobdata"));

    if (m_dfile_item)
        if (DataItem* item = m_dfile_item->dataItem())
            item->setFileName(GUI::Path::intensityDataFileName(batchInfo()->jobName(), "refdata"));
}

void JobItem::initWorker()
{
    ISimulation* simulation =
        GUI::ToCore::itemsToSimulation(*sampleItem(), *instrumentItem(), *simulationOptionsItem())
            .release();
    m_worker = std::make_unique<JobWorker>(batchInfo()->identifier(), simulation);

    m_thread = std::make_unique<QThread>();
    m_worker->moveToThread(m_thread.get());

    connect(worker(), &JobWorker::started, this, &JobItem::onStartedJob);
    connect(worker(), &JobWorker::progressUpdate, this, &JobItem::onItemProgress);
    connect(worker(), &JobWorker::finished, this, &JobItem::onFinishedWork);

    connect(thread(), &QThread::started, worker(), &JobWorker::start);
    connect(thread(), &QThread::finished, this, &JobItem::onFinishedThread);
}

void JobItem::haltWorker()
{
    if (!m_worker)
        return;
    m_worker->terminate();
}

void JobItem::onStartedJob()
{
    batchInfo()->setProgress(0);
    batchInfo()->setStatus(JobStatus::Running);
    batchInfo()->setBeginTime(m_worker->simulationStart());
    batchInfo()->setEndTime(QDateTime());
}

void JobItem::onItemProgress()
{
    if (!worker())
        return;
    batchInfo()->setProgress(worker()->percentageDone());
    emit progressIncremented();
}

void JobItem::onFinishedWork()
{
    ASSERT(worker());
    batchInfo()->setEndTime(worker()->simulationEnd());

    // propagate status of runner
    if (isFailed(worker()->workerStatus()))
        batchInfo()->setComments(worker()->workerFailureMessage());
    else {
        ASSERT(worker()->workerResult());
        simulatedDataItem()->setDatafield(*worker()->workerResult());
        updateFileName();
    }
    batchInfo()->setStatus(worker()->workerStatus());

    // fix job progress (if job was successful, but due to wrong estimation, progress not 100%)
    if (isCompleted(batchInfo()->status()))
        batchInfo()->setProgress(100);

    // tell the thread to exit here (instead of connecting JobRunner::finished
    // to QThread::quit because of strange behaviour)
    ASSERT(thread());
    thread()->quit();

    emit jobFinished(this);

    worker()->disconnect();
    m_worker.release();
}

void JobItem::onFinishedThread()
{
    m_thread->deleteLater();
    m_thread.release();
}
