//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Particle/ReMesocrystal.cpp
//! @brief     Implements class ReMesocrystal.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Particle/ReMesocrystal.h"
#include "Base/Math/Numeric.h"
#include "Base/Spin/SpinMatrix.h"
#include "Base/Types/Span.h"
#include "Base/Vector/WavevectorInfo.h"
#include "Resample/Particle/ReParticle.h"
#include <numbers>

using std::numbers::pi;

ReMesocrystal::ReMesocrystal(const std::optional<size_t>& i_layer, const Lattice3D& lattice,
                             const IReParticle& basis, const ReParticle& outer_shape,
                             const MesoOptions& meso_options, double position_variance)
    : IReParticle(i_layer)
    , m_lattice(lattice)
    , m_basis(basis.clone())
    , m_outer_shape(outer_shape.clone())
    , m_position_variance(position_variance)
    , m_meso_options(meso_options)
{
    if (m_meso_options.use_reciprocal_sum) {
        m_compute_FF = [this](const WavevectorInfo& wv) { return reciprocalSpaceSum(wv); };
        m_compute_FF_pol = [this](const WavevectorInfo& wv) { return reciprocalSpaceSumPol(wv); };
        calculateLargestReciprocalDistance();
    } else {
        m_compute_FF = [this](const WavevectorInfo& wv) { return realSpaceSum(wv); };
        m_compute_FF_pol = [this](const WavevectorInfo& wv) { return realSpaceSumPol(wv); };
    }
}

ReMesocrystal::~ReMesocrystal() = default;

ReMesocrystal* ReMesocrystal::clone() const
{
    return new ReMesocrystal(i_layer(), m_lattice, *m_basis, *m_outer_shape, m_meso_options,
                             m_position_variance);
}

double ReMesocrystal::radialExtension() const
{
    return m_outer_shape->radialExtension();
}

Span ReMesocrystal::zSpan() const
{
    return m_outer_shape->zSpan();
}

complex_t ReMesocrystal::structureFactor(const WavevectorInfo& wavevectors) const
{
    // The underlying implementation computes formfactor exactly, summing over all particles
    // positions inside the mesocrystal shape. Therefore its computational complexity is ~O(N) where
    // 'N' is number of particles in mesocrystal, or ~O(L^3) where 'L' is the linear size of
    // mesocrystal. For mesocrystals larger than ~10x10x10 particles, the approximate algorithm may
    // be preferred.

    C3 q = wavevectors.getQ();

    // The code below is the algorithm to compute the following sum (pseudocode):
    // ***
    // for (R3 position : basis_positions_inside_mesocrystal)
    //     field_factor += exp_I(position.dot(q));
    // ***

    complex_t exp_a = exp_I(m_lattice.basisVectorA().dot(q));
    complex_t exp_b = exp_I(m_lattice.basisVectorB().dot(q));
    complex_t exp_c = exp_I(m_lattice.basisVectorC().dot(q));

    I3 range = m_max_index - m_min_index;

    std::vector<complex_t> exponents_b(range.y() + 1);
    std::vector<complex_t> exponents_c(range.z() + 1);

    for (int j = 0; j <= range.y(); j++)
        exponents_b[j] = pow(exp_b, j + m_min_index.y());
    for (int k = 0; k <= range.z(); k++)
        exponents_c[k] = pow(exp_c, k + m_min_index.z());

    complex_t field_factor(0.0, 0.0);

    //---------------------------------------------------------------------------------

    //    std::vector<complex_t> exponents_a(range.x() + 1);
    //    for (int i = 0; i <= range.x(); i++)
    //        exponents_a[i] = pow(exp_a, i + m_min_index.x());

    //    for (I3 index : m_basis_indexes) {
    //        index -= m_min_index;
    //        field_factor += exponents_a[index.x()] * exponents_b[index.y()] *
    //        exponents_c[index.z()];
    //    }

    //------------------------------------------------------------------

    complex_t denom_factor = Numeric::almostEqual(exp_c, 1., 1) ? 1. : 1. / (exp_c - 1.);
    std::vector<complex_t> exponents_c_r(range.z() + 2);
    for (int k = 0; k <= range.z(); k++)
        exponents_c_r[k] = (pow(exp_c, k + 1) - 1.) * denom_factor;

    if (Numeric::almostEqual(exp_c, 1., 1)) {
        for (int i = 0; i <= range.x(); i++) {
            const complex_t e_a = pow(exp_a, i + m_min_index.x());

            for (int j = 0; j <= range.y(); j++) {
                const complex_t e_ab = e_a * exponents_b[j];

                for (const auto& p : m_pairs[i][j])
                    field_factor += e_ab * complex_t(p.second - p.first + 1);
            }
        }
    } else {
        for (int i = 0; i <= range.x(); i++) {
            const complex_t e_a = pow(exp_a, i + m_min_index.x());

            for (int j = 0; j <= range.y(); j++) {
                const complex_t e_ab = e_a * exponents_b[j];

                for (const auto& p : m_pairs[i][j]) {
                    int N = p.second - p.first;
                    field_factor +=
                        e_ab * exponents_c[p.first - m_min_index.z()] * exponents_c_r[N];
                }
            }
        }
    }

    //------------------------------------------------------------------

    return field_factor * debyeWallerFactor(q.real());
}

complex_t ReMesocrystal::realSpaceSum(const WavevectorInfo& wavevectors) const
{
    return structureFactor(wavevectors) * m_basis->theFF(wavevectors);
}

SpinMatrix ReMesocrystal::realSpaceSumPol(const WavevectorInfo& wavevectors) const
{
    return structureFactor(wavevectors) * m_basis->thePolFF(wavevectors);
}

complex_t ReMesocrystal::reciprocalSpaceSum(const WavevectorInfo& wavevectors) const
{
    // The underlying implementation computes formfactor approximately.
    // The higher the 'm_meso_options.radius_factor', the higher the accuracy.
    // The calculation time is ~O(m_meso_options.radius_factor^3) but it does not depend on the
    // number of particles inside the mesocrystal shape. Therefore, it is reasonable to use this
    // approach for large mesocrystals with thousands of particles inside. The parameter
    // 'm_meso_options.radius_factor' should then be empirically adjusted to some reasonable value.
    // For mesocrystals smaller than ~10x10x10 particles, the exact algorithm may be preferred.

    // retrieve reciprocal lattice vectors within reasonable radius
    C3 q = wavevectors.getQ();
    double radius = m_meso_options.radius_factor * m_max_rec_length;
    std::vector<R3> rec_vectors = m_lattice.reciprocalLatticeVectorsWithinRadius(q.real(), radius);

    // perform convolution on these lattice vectors
    complex_t result(0.0, 0.0);
    for (const auto& rec : rec_vectors) {
        auto dw_factor = debyeWallerFactor(rec);
        WavevectorInfo basis_wavevectors(R3(), -rec, wavevectors.vacuumLambda());
        complex_t basis_factor = m_basis->theFF(basis_wavevectors);
        WavevectorInfo meso_wavevectors(C3(), rec.complex() - q, wavevectors.vacuumLambda());
        complex_t meso_factor = m_outer_shape->theFF(meso_wavevectors);
        result += dw_factor * basis_factor * meso_factor;
    }
    // the transformed delta train gets a factor of (2pi)^3/V, but the (2pi)^3
    // is canceled by the convolution of Fourier transforms :
    return result / m_lattice.unitCellVolume();
}

SpinMatrix ReMesocrystal::reciprocalSpaceSumPol(const WavevectorInfo& wavevectors) const
{
    // See the comment to the non-polarized implementation.

    // retrieve reciprocal lattice vectors within reasonable radius
    C3 q = wavevectors.getQ();
    double radius = m_meso_options.radius_factor * m_max_rec_length;
    std::vector<R3> rec_vectors = m_lattice.reciprocalLatticeVectorsWithinRadius(q.real(), radius);

    // perform convolution on these lattice vectors
    SpinMatrix result;
    for (const auto& rec : rec_vectors) {
        auto dw_factor = debyeWallerFactor(rec);
        WavevectorInfo basis_wavevectors(R3(), -rec, wavevectors.vacuumLambda());
        SpinMatrix basis_factor = m_basis->thePolFF(basis_wavevectors);
        WavevectorInfo meso_wavevectors(C3(), rec.complex() - q, wavevectors.vacuumLambda());
        complex_t meso_factor = m_outer_shape->theFF(meso_wavevectors);
        result += dw_factor * basis_factor * meso_factor;
    }
    // the transformed delta train gets a factor of (2pi)^3/V, but the (2pi)^3
    // is canceled by the convolution of Fourier transforms :
    return result / m_lattice.unitCellVolume();
}

complex_t ReMesocrystal::theFF(const WavevectorInfo& wavevectors) const
{
    return m_compute_FF(wavevectors);
}

SpinMatrix ReMesocrystal::thePolFF(const WavevectorInfo& wavevectors) const
{
    return m_compute_FF_pol(wavevectors);
}

void ReMesocrystal::setBasisIndexes(
    const std::vector<I3>& indexes, const I3& min, const I3& max,
    const std::vector<std::vector<std::vector<std::pair<int, int>>>>& pairs)
{
    m_basis_indexes = indexes;
    m_pairs = pairs;
    m_min_index = min;
    m_max_index = max;
}

void ReMesocrystal::calculateLargestReciprocalDistance()
{
    R3 a1 = m_lattice.basisVectorA();
    R3 a2 = m_lattice.basisVectorB();
    R3 a3 = m_lattice.basisVectorC();

    m_max_rec_length = std::max(pi / a1.mag(), pi / a2.mag());
    m_max_rec_length = std::max(m_max_rec_length, pi / a3.mag());
}

complex_t ReMesocrystal::debyeWallerFactor(const R3& q_i) const
{
    auto q2 = q_i.mag2();
    return std::exp(-q2 * m_position_variance / 2.0);
}
