#include "Base/Axis/Frame.h"
#include "Device/Data/Datafield.h"
#include "GUI/Model/Axis/BasicAxisItem.h"
#include "GUI/Model/Beam/GrazingScanItem.h"
#include "GUI/Model/Beam/SourceItems.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Detector/DetectorItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentsSet.h"
#include "GUI/Model/Files/DatafilesSet.h"
#include "GUI/Model/Project/LinkInstrumentManager.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QSignalSpy>
#include <QTest>

QList<DatafileItem*> linkedRealDataItems(DatafilesSet& realModel,
                                         const InstrumentItem* instrumentItem)
{
    ASSERT(instrumentItem);

    QList<DatafileItem*> result;
    for (auto* dfile_item : realModel.dfileItems()) {
        const QString linkedIdentifier = dfile_item->instrumentId();
        const QString instrumentIdentifier = instrumentItem->id();

        if (linkedIdentifier == instrumentIdentifier)
            result.append(dfile_item);
    }
    return result;
}

TEST(TestLinkInstrument, canLinkToGisas)
{
    ProjectDocument document;

    // populating instrument model
    auto* instrument = document.instrumentsModifier()->addInstrumentItem<GISASInstrumentItem>();
    const QString identifier = instrument->id();
    ASSERT_TRUE(!identifier.isEmpty());

    // populating real data model, setting intensity data
    DatafileItem* realData =
        UTest::GUI::createRealData2D("RealData", *document.datafilesModifier(), 0.);
    const Frame frame = instrument->makeFrame();
    realData->dataItem()->setDatafield({frame.clonedAxes()});

    ASSERT_TRUE(
        document.linkInstrumentManager()->canLinkDataToInstrument(realData, identifier, nullptr));

    // making link
    realData->linkToInstrument(instrument);
    EXPECT_EQ(linkedRealDataItems(*document.datafilesModifier(), instrument), QList<DatafileItem*>()
                                                                                  << realData);

    // changing detector binning and checking that link is destroyed
    auto* detectorItem = instrument->detectorItem();
    detectorItem->phiAxis().setNbins(10);
    emit document.instruments()->instrumentChanged(instrument);

    EXPECT_EQ(linkedRealDataItems(*document.datafilesModifier(), instrument),
              QList<DatafileItem*>());
    EXPECT_EQ(realData->instrumentId(), QString());
}

TEST(TestLinkInstrument, canLinkToSpecular)
{
    ProjectDocument document;

    // populating instrument model
    auto* instrument = dynamic_cast<SpecularInstrumentItem*>(
        document.instrumentsModifier()->addInstrumentItem<SpecularInstrumentItem>());
    const QString identifier = instrument->id();
    ASSERT_TRUE(!identifier.isEmpty());

    // populating real data model, setting intensity data
    DatafileItem* realData =
        UTest::GUI::createRealData1D("RealData", *document.datafilesModifier(), 0.);
    const Frame frame = instrument->makeFrame();
    instrument->scanItem()->grazingScanItem()->initListScan(frame.axis(0));
    instrument->scanItem()->grazingScanItem()->selectListScan();

    realData->dataItem()->setOriginalDatafield({frame.clonedAxes()});

    ASSERT_TRUE(
        document.linkInstrumentManager()->canLinkDataToInstrument(realData, identifier, nullptr));

    // making link
    realData->linkToInstrument(instrument);
    EXPECT_EQ(linkedRealDataItems(*document.datafilesModifier(), instrument), QList<DatafileItem*>()
                                                                                  << realData);
}
